import logging
import typing as typ

import attr

logger = logging.getLogger(__name__)
T = typ.TypeVar("T")


def normalise_name(name: str, upper_names: typ.Collection[str]) -> str:
    if name in upper_names:
        return name
    else:
        if name != name.lower():
            logger.warning(f"The name {name} is not lower case")
        return name.lower()


def normalise_call_names(
    names: typ.Iterable[str], upper_names: typ.Collection[str]
) -> typ.FrozenSet[str]:
    out_names = set()
    for name in names:
        out_names.add(normalise_name(name, upper_names))
    return frozenset(out_names)


@attr.s(frozen=True)
class OrderedCall(typ.Generic[T]):

    name: str = attr.ib()
    call: T = attr.ib(cmp=False, default=lambda info: None)
    requires: typ.FrozenSet[str] = attr.ib(
        cmp=False, default=frozenset(), converter=frozenset
    )
    affects: typ.FrozenSet[str] = attr.ib(
        cmp=False, default=frozenset(), converter=frozenset
    )

    def normalise(self, upper_names: typ.Collection[str] = ()) -> "OrderedCall":
        upper_names = frozenset(upper_names)
        return attr.evolve(
            self,
            name=normalise_name(self.name, upper_names),
            requires=normalise_call_names(self.requires, upper_names),
            affects=normalise_call_names(self.affects, upper_names),
        )

    @classmethod
    def from_callable(cls, call, **k) -> "OrderedCall":
        return cls(name=call.__name__, call=call, **k)


def order_calls_by_dependencies(
    data: typ.Collection[OrderedCall]
) -> typ.Sequence[OrderedCall]:
    tree_inputs: typ.Dict[str, OrderedCall] = {}
    for call in data:
        if call.name in tree_inputs:
            raise ValueError("Cannot add calls with the same name twice")
        tree_inputs[call.name] = call

    requires_set: typ.Dict[str, typ.Set[str]] = {}
    for name, call in tree_inputs.items():
        assert call.requires.issubset(tree_inputs.keys())
        assert call.affects.issubset(tree_inputs.keys())
        requires_set.setdefault(name, set()).update(call.requires)
        for affects_name in call.affects:
            requires_set.setdefault(affects_name, set()).add(name)

    affects_single_set: typ.Dict[str, typ.Set[str]] = {
        name: set() for name in requires_set.keys()
    }
    for after, before_set in requires_set.items():
        for before in before_set:
            affects_single_set[before].add(after)
    affects_combined = {}
    for before, after_set in affects_single_set.items():
        affects_combined[before] = after_set
        affects_combined[before].update(
            *[affects_single_set[after] for after in after_set]
        )
        assert before not in affects_combined[before], f"Cycle detected{requires_set}"

    call_order = sorted(
        affects_single_set.items(),
        key=lambda name_val: (len(name_val[1]), sorted(name_val[1]), name_val[0]),
        reverse=True,
    )
    return tuple(tree_inputs[name] for name, _ in call_order)


if __name__ == "__main__":
    order_calls_by_dependencies(
        {
            OrderedCall("a", requires=["b", "c"]),
            OrderedCall("b", requires=["d", "c"]),
            OrderedCall("c", requires=[]),
            OrderedCall("d", requires=[]),
            OrderedCall("e", requires=[]),
        }
    )



import logging

from matomo_dl.bundle.info import BuildInformation
from matomo_dl.matomo_config import read as read_config, write as write_config

logger = logging.getLogger(__name__)


def update_plugins_list(build: BuildInformation) -> None:
    folder = build.folder
    plugins_folder = folder / "plugins"
    global_config = folder / "config/global.ini.php"
    config = read_config(global_config.open())
    all_plugins = config["Plugins"]["Plugins"]
    default_plugins = config["PluginsInstalled"]["PluginsInstalled"]
    assert isinstance(all_plugins, list)
    assert isinstance(default_plugins, list)
    installed_plugins = list(all_plugins)
    default_plugins = list(default_plugins)
    for plugin_name in all_plugins:
        plugin_folder = plugins_folder / str(plugin_name)
        plugin_json = plugin_folder / "plugin.json"
        if not plugin_folder.exists():
            logger.debug(f"Removing plugin {plugin_name}, not found at {plugin_folder}")
            installed_plugins.remove(plugin_name)
        if not plugin_json.exists():
            logger.debug(
                f"Core plugin {plugin_name}, Loading by default (plugin.json not found at {plugin_json})"
            )
            if plugin_name not in default_plugins:
                default_plugins.append(plugin_name)
    config["Plugins"]["Plugins"] = installed_plugins
    config["PluginsInstalled"]["PluginsInstalled"] = default_plugins
    write_config(config, global_config.open("w"))



from .store import SessionStore

try:
    from requests_cache.core import CachedSession

    class CachingSessionStore(SessionStore, CachedSession):
        pass


except ImportError:
    CachingSessionStore = None



import typing as typ

import attr
import cattr
from packaging.specifiers import SpecifierSet


class Version:

    version: typ.Optional[str] = None
    specifier: SpecifierSet

    @property
    def matches_one_only(self) -> bool:
        return isinstance(self, ExactVersion)

    @property
    def matches_any(self) -> bool:
        return isinstance(self, AnyVersion)

    def choose_version(self, versions: typ.Collection[str]) -> typ.Optional[str]:
        if self.matches_one_only:
            return self.version if self.version in versions else None
        if not versions:
            return None
        assert isinstance(self.specifier, SpecifierSet), repr(self) + repr(
            self.specifier
        )
        matching: typ.List[str] = sorted(self.specifier.filter(versions))
        if not matching:
            return None
        else:
            return matching[-1]

    def __str__(self) -> str:
        return str(self.specifier)


@attr.s
class AnyVersion(Version):
    @property
    def specifier(self) -> SpecifierSet:
        return SpecifierSet("")

    def __str__(self) -> str:
        return "*"


@attr.s
class DynamicVersion(Version):

    specifier: SpecifierSet = attr.ib()


@attr.s
class ExactVersion(Version):

    matches_one_only = True
    version: str = attr.ib()

    @property
    def specifier(self) -> SpecifierSet:
        return SpecifierSet(f"=={self.version}")

    def __str__(self) -> str:
        return self.version


def parse_version(version: str) -> Version:
    version = version.strip()
    if version == "*":
        return AnyVersion()
    if "," in version or version[0] in ["<", ">", "=", "~", "!"]:
        spec_set = SpecifierSet(version)
    else:
        spec_set = SpecifierSet("==" + version)
    if len(spec_set) == 1:
        spec = next(iter(spec_set))
        if spec.operator == "==" and not spec.version.endswith(".*"):
            # Equal, and not a '*' thing. Must be exact.
            return ExactVersion(spec.version)
    return DynamicVersion(spec_set)


def cattr_parse_version(value, _typ) -> Version:
    return parse_version(str(value))


cattr.register_structure_hook(Version, cattr_parse_version)
cattr.register_structure_hook(AnyVersion, cattr_parse_version)
cattr.register_structure_hook(DynamicVersion, cattr_parse_version)
cattr.register_structure_hook(ExactVersion, cattr_parse_version)

cattr.register_unstructure_hook(Version, str)
cattr.register_unstructure_hook(AnyVersion, str)
cattr.register_unstructure_hook(DynamicVersion, str)
cattr.register_unstructure_hook(ExactVersion, str)



import logging
import typing as typ

import attr
import cattr
import toml

from matomo_dl.hashing import HashInfo

logger = logging.getLogger(__name__)


@attr.s
class MatomoLock:

    version: str = attr.ib()
    link: str = attr.ib()
    hash: HashInfo = attr.ib()
    extraction_root: str = attr.ib()


@attr.s
class PluginLock:

    extraction_root: str = attr.ib()


@attr.s
class VersionedPluginLock(PluginLock):

    version: str = attr.ib()
    link: str = attr.ib()
    hash: HashInfo = attr.ib()


@attr.s
class GitPluginLock(PluginLock):

    git: str = attr.ib()
    sha: str = attr.ib()
    hash: HashInfo = attr.ib()


@attr.s
class RawPluginLock(PluginLock):

    link: str = attr.ib()
    hash: HashInfo = attr.ib()


@attr.s
class DistributionLockFile:

    matomo: MatomoLock = attr.ib()
    plugin_locks: typ.Mapping[str, PluginLock] = attr.ib()
    distribution_hash: str = attr.ib()


def unstringify_distribution_lock(dist_lock: str) -> typ.Optional[DistributionLockFile]:
    try:
        data = toml.loads(dist_lock)
        m_lock = cattr.structure(data.pop("matomo"), MatomoLock)
        p_locks = {}
        for plugin_name, lock in data.pop("plugin_locks").items():
            if lock.get("version"):
                p_lock = cattr.structure(lock, VersionedPluginLock)
            elif lock.get("git"):
                p_lock = cattr.structure(lock, GitPluginLock)
            else:
                p_lock = cattr.structure(lock, RawPluginLock)
            p_locks[plugin_name] = p_lock
        data.setdefault("distribution_hash", "")
        return DistributionLockFile(matomo=m_lock, plugin_locks=p_locks, **data)
    except TypeError:
        logger.warning("Failed to load lock file. Discarding contents")
        return None


def stringify_distribution_lock(dist_lock: DistributionLockFile) -> str:
    return toml.dumps(cattr.unstructure(dist_lock))



import difflib
import functools
import itertools
import typing as typ

import click


class ColoredDiffer(difflib.Differ):
    def __init__(self, *a, **k) -> None:
        super().__init__(*a, **k)
        self.line_diff_styles = {
            " ": functools.partial(click.style, dim=True),
            "-": functools.partial(click.style, fg="red"),
            "+": functools.partial(click.style, fg="green"),
        }
        self.char_diff_styles = {
            " ": functools.partial(click.style, dim=True),
            "-": functools.partial(click.style, bg="red"),
            "+": functools.partial(click.style, bg="green"),
            "^": functools.partial(click.style, bold=True),
        }

    def colorise(self, line: str) -> str:
        line = line.rstrip("\n")
        styling = self.line_diff_styles[line[0]]
        tag = styling(line[0], bold=True)
        return f"{tag}{styling(line[1:])}"

    def colorise_char_diff(self, line: str, diff: str) -> typ.Iterator[str]:
        line = line.rstrip("\n")
        diff = diff.rstrip("\n")
        styling = self.line_diff_styles[line[0]]
        out = styling(line[0], bold=True) + " "
        if len(line) >= 3:
            last_char_tag = diff[2]
            run = line[2]
            for char, char_tag in itertools.zip_longest(
                line[3:], diff[3:], fillvalue=" "
            ):
                if last_char_tag != char_tag:
                    out += self.char_diff_styles[last_char_tag](run)
                    run = char
                    last_char_tag = char_tag
                else:
                    run += char
            out += self.char_diff_styles[last_char_tag](run)
        yield out
        # yield diff

    def compare(
        self, a: typ.Union[str, typ.Sequence[str]], b: typ.Union[str, typ.Sequence[str]]
    ) -> typ.Iterator[str]:
        if isinstance(a, str):
            a_seq: typ.Sequence[str] = a.splitlines(keepends=True)
        else:
            a_seq = a
        if isinstance(b, str):
            b_seq: typ.Sequence[str] = b.splitlines(keepends=True)
        else:
            b_seq = b

        last = None
        cmpiter = super().compare(a_seq, b_seq)
        for line in cmpiter:
            tag = line[0]
            if tag != "?":
                if last is not None:
                    yield self.colorise(last)
                last = line
            else:
                assert last is not None
                a_line, a_diff = last, line
                yield from self.colorise_char_diff(a_line, a_diff)
                b_line, b_diff = next(cmpiter), next(cmpiter)
                yield from self.colorise_char_diff(b_line, b_diff)
                last = None



import collections
import re
import typing as typ

SECTION_RE = re.compile(r"^\[(.*?)\](\s*;|$)")
ENTRY_RE = re.compile(r"^(.*?)\s*=\s*(.*?)(\s*;|$)")

ConfigValueScalars = typ.Union[None, int, float, str]
ConfigValues = typ.Union[ConfigValueScalars, typ.List[ConfigValueScalars]]


def read(file: typ.Iterable[str]) -> typ.Dict[str, typ.Dict[str, ConfigValues]]:
    current_section: typ.Dict[str, ConfigValues] = collections.OrderedDict()
    config: typ.Dict[str, typ.Dict[str, ConfigValues]] = collections.OrderedDict()
    for line in file:
        line = line.strip()
        if not line or line[0] == ";":
            continue
        section_match = SECTION_RE.match(line)
        entry_match = ENTRY_RE.match(line)
        if section_match:
            current_section = collections.OrderedDict()
            config[section_match.group(1)] = current_section
        elif entry_match:
            key, value = entry_match.group(1, 2)
            key, value = key.strip(), value.strip()
            if not value:
                real_value: ConfigValueScalars = None
            elif value[0] == value[-1] and value[0] in ['"', "'"]:
                real_value = value[1:-2]
            else:
                try:
                    real_value = int(value)
                except ValueError:
                    try:
                        real_value = float(value)
                    except ValueError:
                        pass
            if key.endswith("[]"):
                val = current_section.setdefault(key[:-2], [])
                assert isinstance(val, list)
                val.append(real_value)
            else:
                current_section[key] = real_value
    return config


def write(config: typ.Dict[str, typ.Dict[str, ConfigValues]], file: typ.IO):
    file.write("; <?php exit; ?> DO NOT REMOVE THIS LINE\n")
    for section, items in config.items():
        file.write(f"\n[{section}]\n")
        file.write("".join(_value_to_string(k, v) for k, v in items.items()))


def _value_to_string(key: str, value: ConfigValues) -> str:
    if value is None:
        return f"{key} =\n"
    elif isinstance(value, str):
        if not value:
            return f"{key} =\n"
        elif value[0] in "0123456789":
            return f'{key} = "{value}"\n'
        else:
            return f"{key} = {value}\n"
    elif isinstance(value, (int, float)):
        return f"{key} = {value}\n"
    else:
        return "".join(_value_to_string(f"{key}[]", val) for val in value)


if __name__ == "__main__":
    import sys

    write(read(open(sys.argv[1])), open(sys.argv[2], "w"))



import logging
import typing as typ

from matomo_dl.distribution.file import GitPlugin, Plugin, VersionedPlugin
from matomo_dl.distribution.lock import GitPluginLock, PluginLock, VersionedPluginLock
from matomo_dl.session import SessionStore
from .git import sync_git_plugin_lock
from .versioned import sync_versioned_plugin_lock

logger = logging.getLogger(__name__)


def sync_plugin_lock(
    session: SessionStore,
    php_version: typ.Optional[str],
    matomo_version: str,
    license_key: typ.Optional[str],
    name: str,
    plugin: Plugin,
    existing_lock: typ.Optional[PluginLock],
) -> PluginLock:
    if isinstance(plugin, VersionedPlugin):
        if not isinstance(existing_lock, VersionedPluginLock):
            existing_lock = None
        return sync_versioned_plugin_lock(
            session,
            php_version,
            matomo_version,
            license_key,
            name,
            plugin.version,
            existing_lock,
        )
    elif isinstance(plugin, GitPlugin):
        if not isinstance(existing_lock, GitPluginLock):
            existing_lock = None
        return sync_git_plugin_lock(
            session,
            php_version,
            matomo_version,
            license_key,
            name,
            plugin.git,
            plugin.ref,
            existing_lock,
        )
    else:
        raise ValueError("Unsupported plugin type.")



import logging
import re
from hashlib import md5

from matomo_dl.bundle.info import BuildInformation

logger = logging.getLogger(__name__)
MANIFEST_ENTRY_RE = re.compile(r'^(\s+)"(.*?)" => .*,$')


def regenerate_manifest(build: BuildInformation) -> None:
    folder = build.folder
    manifest = folder / "config/manifest.inc.php"
    manifest_file = iter(manifest.read_text().splitlines(keepends=False))
    prefix = []
    content = []
    postfix = []
    for line in manifest_file:
        prefix.append(line)
        if "$files" in line:
            break
    for line in manifest_file:
        match = MANIFEST_ENTRY_RE.fullmatch(line)
        if not match:
            postfix.append(line)
            break
        indent, fname = match.groups()
        file = folder / fname
        if not file.exists() or not file.is_file():
            # Remove missing entries
            continue
        stat = file.stat()
        if build.clamp_mtime(stat.st_mtime) == stat.st_mtime:
            # File unchanged in build.
            content.append(line)
            continue
        file_size = stat.st_size
        file_hash = md5(file.read_bytes()).hexdigest()
        content.append(f'{indent}"{fname}" => array("{file_size}", "{file_hash}"),')
    postfix.extend(manifest_file)
    manifest.write_text("\n".join(prefix + sorted(content) + postfix))



import logging
import subprocess
import tempfile
import typing as typ

from matomo_dl.distribution.lock import GitPluginLock
from matomo_dl.errors import MatomoError
from matomo_dl.lock import get_tar_extraction_root
from matomo_dl.session import SessionStore

logger = logging.getLogger(__name__)


def sync_git_plugin_lock(
    session: SessionStore,
    php_version: typ.Optional[str],
    matomo_version: str,
    license_key: typ.Optional[str],
    name: str,
    git_url: str,
    ref: typ.Optional[str],
    existing_lock: typ.Optional[GitPluginLock],
) -> GitPluginLock:
    if not ref:
        ref = "master"
    if existing_lock and ref and existing_lock.sha == ref:
        # Specifying an exact sha & we already locked it.
        return existing_lock
    sha, archive = get_plugin_using_archive(git_url, ref)
    if not sha:
        # Ref may be missing; or the remote server isn't letting us archive a commit sha.
        sha, archive = get_plugin_using_clone(git_url, ref)
    if sha and archive:
        base_path = get_tar_extraction_root(archive, "plugin.json")
        if not base_path:
            logger.error("Cannot determine how to extract plugin!")
            raise ValueError("Unable to determine path")
        archive_hash = session.store_cache_data(f"git-plugin-{name}-{sha}-tar", archive)
        return GitPluginLock(
            git=git_url, extraction_root=base_path, sha=sha, hash=archive_hash
        )
    else:
        logger.warning("Cannot retrieve {ref!r} fron {git_url!r}.")
        raise MatomoError("Cannot create an archive of {ref!r}. Does it exist?")


def get_plugin_using_archive(
    git_url: str, ref: str
) -> typ.Tuple[typ.Optional[str], typ.Optional[bytes]]:
    try:
        with tempfile.NamedTemporaryFile(suffix=".tar", mode="rb") as tarball:
            subprocess.run(
                [
                    "git",
                    "archive",
                    "--format=tar",
                    "--output",
                    tarball.name,
                    "--remote",
                    git_url,
                    ref,
                ],
                check=True,
            )
            with open(tarball.name, "rb") as f:
                # Prefer 'stdin' over 'input' as this command only uses the first part of the file
                res = subprocess.run(
                    ["git", "get-tar-commit-id"],
                    stdin=f,
                    universal_newlines=True,
                    stdout=subprocess.PIPE,
                    check=True,
                )
            return res.stdout.strip(), tarball.read()
    except subprocess.CalledProcessError:
        return None, None


def get_plugin_using_clone(
    git_url: str, ref: str
) -> typ.Tuple[typ.Optional[str], typ.Optional[bytes]]:
    try:
        with tempfile.TemporaryDirectory() as gitdir:
            subprocess.run(
                ["git", "clone", "--depth=5", "--bare", git_url, gitdir], check=True
            )
            res = subprocess.run(
                ["git", "rev-parse", f"{ref}^{{commit}}"],
                cwd=gitdir,
                universal_newlines=True,
                stdout=subprocess.PIPE,
            )
            if res.returncode != 0:
                subprocess.run(
                    ["git", "fetch", "--unshallow", "origin"], cwd=gitdir, check=True
                )
                res = subprocess.run(
                    ["git", "rev-parse", f"{ref}^{{commit}}"],
                    cwd=gitdir,
                    universal_newlines=True,
                    check=True,
                    stdout=subprocess.PIPE,
                )
            sha = res.stdout.strip()
            with tempfile.NamedTemporaryFile(suffix=".tar", mode="rb") as tarball:
                subprocess.run(
                    [
                        "git",
                        "archive",
                        "--format=tar",
                        "--output",
                        tarball.name,
                        "--remote",
                        sha,
                    ],
                    cwd=gitdir,
                    check=True,
                )
                return sha, tarball.read()

    except subprocess.CalledProcessError:
        return None, None



import logging
import typing as typ

from matomo_dl.distribution.file import DistributionFile
from matomo_dl.distribution.lock import DistributionLockFile
from matomo_dl.lock.matomo import sync_matomo_lock
from matomo_dl.lock.plugin import sync_plugin_lock
from matomo_dl.progress import progressbar
from matomo_dl.session import SessionStore

logger = logging.getLogger(__name__)
DEFAULT_ASSUMED_PHP_VERSION = "7.2"


def build_lock(
    session: SessionStore,
    dist: DistributionFile,
    lock: typ.Optional[DistributionLockFile],
) -> DistributionLockFile:
    with progressbar(label="Syncing lock files", length=len(dist.plugins) + 1) as bar:
        matomo_lock = sync_matomo_lock(
            session, dist.version, lock.matomo if lock else None
        )
        bar.update(1)
        if lock:
            old_plugin_locks = {
                name: plugin_lock for name, plugin_lock in lock.plugin_locks.items()
            }
        else:
            old_plugin_locks = {}
        plugin_locks = {}
        for i, (name, plugin) in enumerate(dist.plugins.items(), start=1):
            bar.update(i)
            p_lock = sync_plugin_lock(
                session,
                dist.php_version,
                matomo_lock.version,
                dist.license_key,
                name,
                plugin,
                old_plugin_locks.get(name),
            )
            plugin_locks[name] = p_lock
    return DistributionLockFile(
        matomo=matomo_lock,
        plugin_locks=plugin_locks,
        distribution_hash=dist.versioning_hash,
    )



import logging
import pathlib
import typing as typ
from collections import defaultdict
from hashlib import md5
from types import SimpleNamespace

from matomo_dl.bundle.fs_util import iter_tree
from matomo_dl.bundle.info import BuildInformation

logger = logging.getLogger(__name__)


def list_duplicates(build: BuildInformation) -> None:
    files_by_size: typ.Dict[int, typ.Set[pathlib.Path]] = defaultdict(set)
    for file in iter_tree(build.folder):
        if file.is_file():
            files_by_size[file.stat().st_size].add(file)
    for size, file_group in files_by_size.items():
        if size == 0 or len(file_group) < 2:
            continue
        files_by_hash: typ.Dict[str, typ.Set[pathlib.Path]] = defaultdict(set)
        for file in file_group:
            files_by_hash[md5(file.read_bytes()).hexdigest()].add(file)
        for matching_files in files_by_hash.values():
            if len(matching_files) < 2:
                continue
            else:
                print(matching_files)


if __name__ == "__main__":
    d = SimpleNamespace()
    d.folder = pathlib.Path(".")
    list_duplicates(d)  # type: ignore



import json
import os
import pathlib
import typing as typ
from hashlib import sha256

import attr
import cattr
import toml

from .customisations import Customisations
from .version import AnyVersion, Version


class Plugin:
    pass


@attr.s
class VersionedPlugin(Plugin):

    version: Version = attr.ib()


@attr.s
class GitPlugin(Plugin):

    git: str = attr.ib()
    ref: typ.Optional[str] = attr.ib()


@attr.s
class RawPlugin(Plugin):

    link: str = attr.ib()


@attr.s
class DistributionFile:

    version: Version = attr.ib()
    php_version: typ.Optional[str] = attr.ib(default=None)
    license_key: typ.Optional[str] = attr.ib(default=None)
    plugins: typ.Mapping[str, Plugin] = attr.ib(factory=dict)
    customisation: Customisations = attr.ib(factory=Customisations)

    @classmethod
    def _normalise_license_key(
        cls, base_dir: typ.Optional[pathlib.Path], key: typ.Optional[str]
    ) -> typ.Optional[str]:
        if key is None:
            return None
        key = key.strip()
        if not key:
            return None
        elif key[0] == "$":
            return cls._normalise_license_key(base_dir, os.environ[key[1:]])
        elif key[0] == "<":
            if base_dir is None:
                raise ValueError("Cannot normalise a file include without a base path")
            return cls._normalise_license_key(
                base_dir, (base_dir / key[1:]).read_text()
            )
        else:
            return key

    def normalise_license_key(self, base_dir: typ.Optional[pathlib.Path]):
        self.license_key = self._normalise_license_key(base_dir, self.license_key)

    @property
    def versioning_hash(self) -> str:
        versioning_json = json.dumps(
            {
                "version": cattr.unstructure(self.version),
                "php_version": cattr.unstructure(self.php_version),
                "plugins": cattr.unstructure(self.plugins),
            },
            indent=None,
            separators=(",", ":"),
            sort_keys=True,
        )
        return "sha256:" + sha256(versioning_json.encode("utf-8")).hexdigest()


def parse_plugin(value, _typ) -> Plugin:
    if value is None:
        return VersionedPlugin(AnyVersion())
    elif isinstance(value, str):
        return VersionedPlugin(cattr.structure(value, Version))
    elif not isinstance(value, dict):
        raise ValueError("Unsupported plugin type {value!r}")
    elif "version" in value:
        value = value.pop("version")
        if value is None:
            return VersionedPlugin(AnyVersion())
        elif isinstance(value, str):
            return VersionedPlugin(cattr.structure(value, Version))
    elif "git" in value:
        return GitPlugin(**value)
    elif "link" in value:
        return RawPlugin(**value)
    raise ValueError("Unsupported plugin type {value!r}")


cattr.register_structure_hook(Plugin, parse_plugin)


def unstringify_distribution_file(
    base_dir: pathlib.Path, file_content: str
) -> DistributionFile:
    v: DistributionFile = cattr.structure(toml.loads(file_content), DistributionFile)
    v.normalise_license_key(base_dir)
    return v


def stringify_distribution_file(dist_file: DistributionFile) -> str:
    return toml.dumps(cattr.unstructure(dist_file))



import logging

from matomo_dl.bundle.fs_util import delete_all_matching
from matomo_dl.bundle.info import BuildInformation

logger = logging.getLogger(__name__)


def remove_example_plugins(build: BuildInformation) -> None:
    build.add_removed_files(
        delete_all_matching(
            build.folder,
            folders=[
                "plugins/ExampleAPI",
                "plugins/ExampleCommand",
                "plugins/ExamplePlugin",
                "plugins/ExampleReport",
                "plugins/ExampleSettingsPlugin",
                "plugins/ExampleTheme",
                "plugins/ExampleTracker",
                "plugins/ExampleUI",
                "plugins/ExampleVisualization",
            ],
        )
    )


def remove_build_support(build: BuildInformation) -> None:
    build.add_removed_files(
        delete_all_matching(
            build.folder,
            extensions=[".gzip", ".php4", ".feature"],
            names=[
                ".codeclimate.yml",
                ".htaccess",
                ".lfsconfig",
                ".npmignore",
                ".php_cs.dist",
                ".scrutinizer.yml",
                "behat.yml",
                "bower.json",
                "build.properties",
                "build.xml",
                "component.json",
                "composer.json",
                "composer.lock",
                "composer.travis.json",
                "couscous.yml",
                "grumphp.yml",
                "gruntfile.js",
                "installed.json",
                "karma.conf.js",
                "makefile",
                "package.json",
                "package.xml",
                "phpbench.json",
                "phpunit.xml.dist",
                "phpunit.xml",
                "protractor.conf.js",
            ],
            regexes=[r"libs/bower_components/jquery-ui/ui/jquery-ui[a-z\-.]+js"],
            folders=[
                "libs/bower_components/angular-mocks/",
                "libs/bower_components/jquery-ui/ui/i18n/",
                "libs/bower_components/sprintf/demo/",
            ],
            paths=[
                "libs/bower_components/iframe-resizer/test-main.js",
                "libs/bower_components/jScrollPane/script/demo.js",
                "libs/bower_components/jScrollPane/style/demo.css",
                "libs/bower_components/materialize/package.js",
                "libs/bower_components/ngDialog/server.js",
                "libs/bower_components/visibilityjs/index.js",
                "libs/bower_components/visibilityjs/logo.svg",
                "libs/jqplot/build_minified_script.sh",
                "plugins/AbTesting/libs/jquery-timepicker/jt.timepicker.jquery.json",
                "plugins/AbTesting/redirect/vendor/innocraft/php-experiments/docs/generateDocs.sh",
                "vendor/leafo/lessphp/lessify",
                "vendor/leafo/lessphp/package.sh",
                "vendor/leafo/lessphp/plessc",
                "vendor/pear/archive_tar/scripts/phptar.in",
                "vendor/pear/archive_tar/sync-php4",
            ],
        )
    )


def remove_documentation(build: BuildInformation) -> None:
    build.add_removed_files(
        delete_all_matching(
            build.folder,
            extensions=[".md", ".rst", ".markdown", ".log"],
            names=["license", "license.txt"],
            stems=[
                "authors",
                "changelog",
                "gnu-lgpl",
                "gpl-2.0",
                "copying",
                "gpl-3.0",
                "legalnotice",
                "license-colors",
                "license-sizzle",
                # "license",  # TODO: figure out how to keep 'plugins/Marketplace/angularjs/licensekey/'
                "mit and gpl2 licenses",
                "mit-license-history",
                "mit-license",
                "readme",
            ],
            folders=[
                "misc/composer/",
                "misc/cron/",
                "misc/others/",
                "misc/proxy-hide-piwik-url/",
            ],
            paths=[
                "misc/user/index.html",
                "misc/How to install Matomo.html",
                "vendor/pear/archive_tar/docs/Archive_Tar.txt",
                "libs/bower_components/chroma-js/LICENSE-colors",
            ],
        )
    )


def remove_vendored_extras(build: BuildInformation) -> None:
    build.add_removed_files(
        delete_all_matching(
            build.folder,
            folders=["vendor/tecnickcom/tcpdf/tools/", "vendor/twig/twig/ext/"],
        )
    )



import logging
import os
import pathlib
import time
import typing as typ

import attr
import cattr

from matomo_dl.distribution.lock import DistributionLockFile

logger = logging.getLogger(__name__)
BUILD_START_TIME = round(time.time())


def get_environ_build_mtime() -> typ.Optional[int]:
    if os.environ.get("SOURCE_DATE_EPOCH", None):
        try:
            return int(os.environ["SOURCE_DATE_EPOCH"])
        except ValueError:
            raise ValueError(
                "The value in $SOURCE_DATE_EPOCH is not a parsable integer"
            )
    return None


def get_build_mtime(default: typ.Optional[int] = None) -> int:
    env_mtime = get_environ_build_mtime()
    if env_mtime is None:
        if default is None:
            return BUILD_START_TIME
        else:
            return default
    else:
        return env_mtime


@attr.s
class BuildInformation:

    lockfile: DistributionLockFile = attr.ib()
    customisations: "Customisations" = attr.ib()
    folder: pathlib.Path = attr.ib()
    extra_info: typ.Dict[str, typ.Any] = attr.ib(factory=dict)
    mtime_clamp: int = attr.ib(factory=get_build_mtime)

    def add_source_time(self, source: int):
        if source > BUILD_START_TIME:
            logger.warning("Source built after build started.")
        elif self.mtime_clamp == BUILD_START_TIME:
            self.mtime_clamp = source
        else:
            self.mtime_clamp = max(self.mtime_clamp, source + 1)

    def clamp_mtime(self, source: typ.Union[int, float]) -> int:
        return int(min(source, self.mtime_clamp))

    def to_output(self) -> typ.Dict[str, typ.Any]:
        data: typ.Dict[str, typ.Any] = cattr.unstructure(self)
        del data["folder"]
        return data

    def add_removed_files(self, files: typ.Iterable[typ.Union[pathlib.Path, str]]):
        clean_files: typ.Set[str] = set()
        if "removed_files" in self.extra_info:
            clean_files.update(self.extra_info["removed_files"])
        for file in files:
            if isinstance(file, pathlib.Path):
                if file.is_absolute():
                    clean_files.add(str(file.relative_to(self.folder)))
                else:
                    clean_files.add(str(file))
            else:
                clean_files.add(str(file))
        self.extra_info["removed_files"] = sorted(clean_files)


if typ.TYPE_CHECKING:
    # At the end to avoid a recursive import
    from matomo_dl.distribution.customisations import Customisations
else:
    Customisations = None



import pathlib
import re
import typing as typ

import requests
from requests.adapters import HTTPAdapter

from matomo_dl.hashing import HashInfo, all_hashes_for_data


class HttpLoggingAdapter(HTTPAdapter):
    def send(self, request, *a, **k):
        if (
            isinstance(request, requests.Request)
            and request.url
            and request.url.startswith("http://")
        ):
            print("WARNING: Request sent over HTTP.\n\t{}".format(request.url))
        return super().send(request, *a, **k)


class SessionStore(requests.Session):
    cache_dir: typ.Optional[pathlib.Path]

    def __init__(
        self, *a: typ.Any, cache_dir: typ.Optional[pathlib.Path], **k: typ.Any
    ):
        super().__init__(*a, **k)  # type: ignore
        self.mount("http://", HttpLoggingAdapter())
        if cache_dir:
            self.cache_dir = pathlib.Path(cache_dir)
        else:
            self.cache_dir = None

    def store_cache_data(self, cache_key: str, data: bytes) -> HashInfo:
        hashes = all_hashes_for_data(data)
        if self.cache_dir:
            assert re.match(r"^[0-9a-z\-_.]+$", cache_key)
            file = self.cache_dir / "{}.dat".format(cache_key)
            file.write_bytes(data)
            hash_file = self.cache_dir / "{}.dat.check".format(cache_key)
            hash_file.write_text(hashes)
        return hashes

    def retrieve_cache_data(
        self, cache_key: str, expected_hash: HashInfo
    ) -> typ.Optional[bytes]:
        if not self.cache_dir:
            return None
        assert re.match(r"^[0-9a-z\-_.]+$", cache_key)
        assert expected_hash
        file = self.cache_dir / "{}.dat".format(cache_key)
        hash_file = self.cache_dir / "{}.dat.check".format(cache_key)
        if not file.exists() or not hash_file.exists():
            return None
        digests = hash_file.read_text().splitlines()
        if expected_hash in digests:
            data = file.read_bytes()
            if expected_hash == all_hashes_for_data(data):
                return data

        file.unlink()
        hash_file.unlink()
        return None



import gzip
import json
import logging
import pathlib
import tarfile
import tempfile
import typing as typ
from contextlib import ExitStack

from matomo_dl.distribution.file import DistributionFile
from matomo_dl.distribution.lock import DistributionLockFile
from matomo_dl.errors import DownloadHashMismatch
from matomo_dl.lock.matomo import get_cache_key
from matomo_dl.progress import progressbar
from matomo_dl.session import SessionStore
from .customisation import apply_customisations
from .extract import extract_zip_file
from .info import BuildInformation
from .plugin import extract_plugins
from .stat import standardise_mode

logger = logging.getLogger(__name__)


def build_release(
    session: SessionStore,
    dist: DistributionFile,
    lock: DistributionLockFile,
    output_file: pathlib.Path,
):
    with tempfile.TemporaryDirectory(prefix="matomo-dl") as f:
        folder = pathlib.Path(f)
        info = BuildInformation(lock, dist.customisation, folder)
        extract_matomo(session, info)
        extract_plugins(session, dist.license_key, info)

        apply_customisations(info)

        (folder / ".build.json").write_text(json.dumps(info.to_output()))
        create_release_tarball(info, output_file)


def extract_matomo(session: SessionStore, build: BuildInformation):
    with progressbar(range(4), label="Acquiring Matomo") as _bar:
        logger.info("Attempting to retrieve Matomo zipfile from cache")
        bar = iter(_bar)
        lock = build.lockfile.matomo
        folder = build.folder
        next(bar)
        data = session.retrieve_cache_data(get_cache_key(lock.version), lock.hash)
        next(bar)
        if not data:
            logger.info("Matomo zipfile not in cache. Downloading it now")
            r = session.get(lock.link)
            next(bar)
            data = r.content
            data_hash = session.store_cache_data(get_cache_key(lock.version), data)
            next(bar)
            if lock.hash != data_hash:
                raise DownloadHashMismatch()
        list(bar)  # Finish the progress bar
    latest_mtime = extract_zip_file(
        data, folder, root=lock.extraction_root, progress="Extracting Matomo"
    )
    assert latest_mtime is not None
    build.add_source_time(latest_mtime)


def create_release_tarball(build: BuildInformation, output_file: pathlib.Path):
    if output_file.suffixes[0] != ".tar":
        output_file = output_file.with_suffix(".tar.gz")
    with ExitStack() as stack:
        if output_file.suffix == ".tar":
            file: tarfile.TarFile = stack.enter_context(
                tarfile.open(output_file, mode="w")
            )
        elif output_file.suffix == ".gz":
            # Have to construct the gzip stream seperately otherwise python
            #  'helpfully' adds the current time as a gzip comment; thus
            #  ruining our reproducible builds.
            output_fd: typ.IO[bytes] = gzip.GzipFile(  # type: ignore
                filename="",
                mode="wb",
                compresslevel=9,
                mtime=build.mtime_clamp,
                fileobj=stack.enter_context(output_file.open("wb")),
            )
            file = stack.enter_context(
                tarfile.open(fileobj=stack.enter_context(output_fd), mode="w")
            )
        else:
            if output_file.suffix in [".gz", ".xz"]:
                mode = "w:" + output_file.suffix[1:]
            elif output_file.suffix in [".bz", ".bz2"]:
                mode = "w:bz2"
            else:
                raise ValueError()
            file = stack.enter_context(
                tarfile.open(output_file, mode=mode, compresslevel=9)
            )
        items = list(iter_folder_for_tar(file, build, build.folder, ""))
        bar = stack.enter_context(progressbar(items, label="Creating archive"))
        for tar_info, path in bar:
            if path:
                with path.open("rb") as f:
                    file.addfile(tar_info, f)
            else:
                file.addfile(tar_info)


def normalise_tar_info(
    build: BuildInformation, info: tarfile.TarInfo
) -> typ.Optional[tarfile.TarInfo]:
    if not info.isfile() and not info.isdir():
        return None
    if info.mode & 0o7000 != 0:
        print(info, info.mode)
        raise ValueError()
    info.mode = standardise_mode(info.mode, force_exec=info.isdir())
    info.uid = 0
    info.uname = ""
    info.gid = 0
    info.gname = ""
    info.pax_headers = {}
    # Clamp any files modified during this run to the start;
    # See https://reproducible-builds.org/docs/source-date-epoch/
    info.mtime = build.clamp_mtime(info.mtime)
    return info


def iter_folder_for_tar(
    tar: tarfile.TarFile, build: BuildInformation, folder: pathlib.Path, base_path: str
) -> typ.Iterator[typ.Tuple[tarfile.TarInfo, typ.Optional[pathlib.Path]]]:
    if base_path != "" and base_path[-1] != "/":
        base_path += "/"
    for path in sorted(folder.iterdir()):
        tar_info = normalise_tar_info(
            build, tar.gettarinfo(str(path), f"{base_path}{path.name}")
        )
        if not tar_info:
            continue
        elif tar_info.isfile():
            yield tar_info, path
        elif tar_info.isdir():
            yield tar_info, None
            yield from iter_folder_for_tar(tar, build, path, f"{base_path}{path.name}/")



import itertools
import logging
import pathlib
import re
import typing as typ

logger = logging.getLogger(__name__)


def iter_tree(folder: pathlib.Path) -> typ.Iterator[pathlib.Path]:
    for path in folder.iterdir():
        if path.is_file():
            yield path
        elif path.is_dir():
            yield path
            yield from iter_tree(path)


def get_path_regex(
    base_path,
    *,
    regexes: typ.Iterable[str] = (),
    folders: typ.Iterable[str] = (),
    paths: typ.Iterable[str] = (),
    names: typ.Iterable[str] = (),
    stems: typ.Iterable[str] = (),
    extensions: typ.Iterable[str] = (),
) -> typ.Pattern:
    names_re = or_regex(map(re.escape, names))
    stems_re = or_regex(map(re.escape, stems))
    extensions_re = or_regex(map(re.escape, extensions))
    folders_re = or_regex(map(re.escape, map(lambda s: s.rstrip("/"), folders)))
    resp = list(regexes)
    resp.extend(map(re.escape, paths))
    if names_re:
        resp.append(f".*/{names_re}")
    if extensions_re:
        resp.append(f".*{extensions_re}")
    if stems_re:
        resp.append(f".*/{stems_re}(?:.[^/.]+)?$")
    if folders_re:
        resp.append(f"{folders_re}/.*")
    resp_re = or_regex(resp)
    assert resp_re
    return re.compile("^" + re.escape(str(base_path)) + resp_re + "$", re.IGNORECASE)


def or_regex(res: typ.Iterable[str]) -> typ.Optional[str]:
    regs = set(res)
    if len(regs) == 0:
        return None
    elif len(regs) == 1:
        return regs.pop()
    else:
        return f"(?:{'|'.join(regs)})"


def iter_all_matching(
    path: pathlib.Path, pattern: typ.Pattern
) -> typ.Iterator[pathlib.Path]:
    return filter(lambda path: pattern.match(str(path)), iter_tree(path))


def delete_all_matching(
    path: pathlib.Path, *, folders=(), paths=(), **kwds
) -> typ.Iterator[pathlib.Path]:
    chainable = [
        delete_all_iter(path / file for file in set(paths)),
        delete_all_iter(path / folder for folder in set(folders)),
    ]
    if kwds:
        chainable.append(
            delete_all_iter(iter_all_matching(path, get_path_regex(path, **kwds)))
        )
    return itertools.chain.from_iterable(chainable)


def delete_all(*paths: pathlib.Path) -> typ.Collection[pathlib.Path]:
    return delete_all_iter(paths)


def delete_all_iter(paths: typ.Iterable[pathlib.Path]) -> typ.Collection[pathlib.Path]:
    to_delete = set()
    for item in paths:
        to_delete.add(item)
    deleted = set()
    while to_delete:
        item = to_delete.pop()
        if item.is_dir():
            folder_iter = iter(item.iterdir())
            first = next(folder_iter, None)
            if first is None:
                item.rmdir()
                deleted.add(item)
            else:
                to_delete.update(folder_iter)
                to_delete.add(item)
                to_delete.add(first)
        elif item.exists():
            item.unlink()
            deleted.add(item)
    return deleted



import logging
import os
import pathlib
import typing as typ
import zipfile
from contextlib import ExitStack
from datetime import datetime
from io import BytesIO

from matomo_dl.progress import progressbar
from .stat import is_extended_mode_dir, standardise_privacy_mode

logger = logging.getLogger(__name__)


def is_zipinfo_dir(info: zipfile.ZipInfo) -> bool:
    is_msdos_dir = bool(info.external_attr & 0x10)
    is_unix_dir = bool(is_extended_mode_dir(info.external_attr >> 16))
    return is_msdos_dir or is_unix_dir or info.filename[-1] == "/"


def resolve_filename(
    destination: pathlib.Path, name: str
) -> typ.Optional[pathlib.Path]:
    target = (destination / name).resolve(strict=False)
    if (
        len(target.parts) <= len(destination.parts)
        or target.parts[: len(destination.parts)] != destination.parts
    ):
        return None
    return target


def extract_zip_file(
    file_data: bytes,
    destination: pathlib.Path,
    root: str,
    progress: typ.Optional[str] = None,
) -> typ.Optional[int]:
    latest_mtime = None
    with ExitStack() as stack:
        file = stack.enter_context(zipfile.ZipFile(BytesIO(file_data), "r"))
        infos: typ.Iterable[zipfile.ZipInfo] = file.infolist()
        if progress:
            infos = stack.enter_context(progressbar(infos, label=progress))
        for item in infos:
            _, root_chk, dest_filename = item.filename.partition(root)
            if root_chk != root:
                continue
            elif not dest_filename:
                destination.mkdir(parents=True, exist_ok=True)
                continue
            dest = resolve_filename(destination, dest_filename)
            if not dest:
                logger.warning(
                    f"Skipping potentially dangeous item {item.filename!r}, "
                    f"it was to be placed at {dest_filename!r}"
                )
                continue
            if is_zipinfo_dir(item):
                dest.mkdir(parents=True, exist_ok=True)
            else:
                mtime = int(datetime(*item.date_time, microsecond=0).timestamp())
                if latest_mtime is None:
                    latest_mtime = mtime
                else:
                    latest_mtime = max(latest_mtime, mtime)
                # File
                mode = standardise_privacy_mode(item.external_attr >> 16)
                dest.parent.mkdir(parents=True, exist_ok=True)
                dest.write_bytes(file.read(item))
                dest.chmod(mode)
                os.utime(dest, times=(mtime, mtime))
    return latest_mtime



import logging
import typing as typ
from urllib.parse import urljoin

import attr
import requests

from matomo_dl.distribution.lock import VersionedPluginLock
from matomo_dl.distribution.version import Version, parse_version
from matomo_dl.errors import MissingDownloadError, VersionError
from matomo_dl.lock import get_zip_extraction_root
from matomo_dl.session import SessionStore

logger = logging.getLogger(__name__)
PLUGIN_API_URL = "https://plugins.matomo.org"


def get_cache_key(name: str, version: str) -> str:
    return f"plugin-{name.lower()}-{version}-zip"


def sync_versioned_plugin_lock(
    session: SessionStore,
    php_version: typ.Optional[str],
    matomo_version: str,
    license_key: typ.Optional[str],
    name: str,
    version_spec: Version,
    existing_lock: typ.Optional[VersionedPluginLock],
) -> VersionedPluginLock:
    version, dl_url = resolve_plugin_version_spec(
        session, php_version, matomo_version, license_key, version_spec, name
    )
    if existing_lock and version == existing_lock.version:
        return existing_lock

    cache_key = get_cache_key(name, version)
    resp = plugin_request(session, dl_url, license_key=license_key)
    resp.raise_for_status()
    data = resp.content

    base_path = get_zip_extraction_root(data, "plugin.json")
    if not base_path:
        logger.error("Cannot determine how to extract plugin!")
        raise ValueError("Unable to determine path")
    hashes = session.store_cache_data(cache_key, data)
    return VersionedPluginLock(
        version=version, link=dl_url, extraction_root=base_path, hash=hashes
    )


def resolve_plugin_version_spec(
    session: requests.Session,
    php_version: typ.Optional[str],
    matomo_version: str,
    license_key: typ.Optional[str],
    version_spec: Version,
    name: str,
) -> typ.Tuple[str, str]:
    latest, all_versions = get_all_plugin_versions(
        session, php_version, matomo_version, license_key, name
    )
    if not latest and not all_versions:
        logger.error(f"There are no versions of {name} that are supported.")
        raise VersionError(f"No versions returned for {name}")
    elif latest and version_spec.choose_version([latest]):
        return latest, all_versions[latest]
    else:
        version = version_spec.choose_version(set(all_versions))
        if version:
            return version, all_versions[version]
        else:
            logger.error(
                f"There are no versions of {name} that "
                f"match the version specifier {version_spec}."
            )
            raise VersionError(f"No matching versions for {name}")


@attr.s
class PluginData:

    name: str = attr.ib()
    is_downloadable: bool = attr.ib()
    is_paid: bool = attr.ib()
    versions: typ.Collection[typ.Mapping[str, typ.Any]] = attr.ib()
    latest_version: typ.Optional[str] = attr.ib()
    raw_response: requests.Response = attr.ib()

    def assert_downloadable(self) -> None:
        if self.is_downloadable:
            return
        if self.is_paid:
            raise MissingDownloadError(
                f"The plugin {self.name!r} is a paid-for plugin. Ensure a license "
                f"key is provided which has purchased this plugin"
            )
        else:
            raise MissingDownloadError(
                f"Cannot download {self.name!r} with current configuration."
            )

    def iter_versions(
        self,
    ) -> typ.Iterator[typ.Tuple[str, str, typ.Mapping[str, str]]]:
        for version in self.versions:
            ver = version["name"]
            dl_url = urljoin(self.raw_response.url, version["download"])
            requires = version.get("requires", {})
            yield (ver, dl_url, requires)


def get_plugin_data(
    session: requests.Session,
    license_key: typ.Optional[str],
    name: str,
    matomo_version: typ.Optional[str] = None,
) -> PluginData:
    url = f"{PLUGIN_API_URL}/api/2.0/plugins/{name}/info"
    if matomo_version:
        url += f"?coreVersion={matomo_version}"
    resp = plugin_request(session, url, license_key=license_key)
    resp.raise_for_status()
    data = resp.json()
    return PluginData(
        name=data.get("name"),
        is_downloadable=data.get("isDownloadable", False),
        is_paid=data.get("isPaid", False),
        versions=data.get("versions", ()),
        latest_version=data.get("latestVersion", None),
        raw_response=resp,
    )


def get_all_plugin_versions(
    session: requests.Session,
    php_version: typ.Optional[str],
    matomo_version: str,
    license_key: typ.Optional[str],
    name: str,
) -> typ.Tuple[typ.Optional[str], typ.Mapping[str, str]]:
    data = get_plugin_data(session, license_key, name, matomo_version)
    data.assert_downloadable()
    latest_version = data.latest_version
    filtered_versions = {}
    for name, dl_url, requires in data.iter_versions():
        if "piwik" in requires:
            matomo = parse_version(requires["piwik"])
            if not matomo.choose_version([matomo_version]):
                continue
        if php_version and "php" in requires:
            php = parse_version(requires["php"])
            if not php.choose_version([php_version]):
                continue
        filtered_versions[name] = dl_url
    if not filtered_versions:
        raise VersionError(
            f"There are no versions of {name!r} supported on"
            f" Matomo {matomo_version} and PHP {php_version}."
        )
        return (None, {})
    if latest_version not in filtered_versions:
        logger.warning(
            f"The latest version of {name!r} is not supported on"
            f" Matomo {matomo_version} and PHP {php_version}."
        )
        return None, filtered_versions
    else:
        return latest_version, filtered_versions


def plugin_request(
    session: requests.Session, *a, license_key: typ.Optional[str] = None, **k
) -> requests.Response:
    if not license_key:
        return session.post(*a, **k)
    data = None
    if "data" not in k:
        k["data"] = {"access_token": license_key}
    elif isinstance(k["data"], typ.MutableMapping):
        k["data"].setdefault("access_token", license_key)
    resp = session.post(*a, **k)
    if resp.status_code in [401, 403]:
        logger.warning("License key denied. Contact Matomo for support.")
        if "data" in k and isinstance(k["data"], typ.MutableMapping):
            k["data"].pop("access_token", None)
        return plugin_request(*a, **k, data=data)
    return resp



import functools
import hashlib
import typing as typ

HashInfo = str


desired_algorithms = ["sha256"]


def all_hashes_for_data(data: bytes) -> HashInfo:
    return ":".join(next(hashes_for_data(data)))


def hashes_for_data(data: bytes) -> typ.Iterator[typ.Tuple[str, str]]:
    hashed = False
    for algo in desired_algorithms:
        if algo in hashlib.algorithms_available:
            hashed = True
            yield (algo, hashlib.new(algo, data).hexdigest())
    assert hashed


@functools.lru_cache()
def hash_for_data(data: bytes, algo: str) -> str:
    assert algo in hashlib.algorithms_available
    return hashlib.new(algo, data).hexdigest()



import stat

user_execute_bit = stat.S_IXUSR


def is_mode_executable(mode: int) -> bool:
    return (stat.S_IMODE(mode) & 0o111) != 0


def standardise_mode(mode: int, *, force_exec: bool = False) -> int:
    if force_exec or is_mode_executable(mode):
        # User has full read/write/execute. Others have read/execute
        return 0o755
    else:
        # User has full read/write. Others have read
        return 0o644


def standardise_privacy_mode(mode: int, *, force_exec: bool = False) -> int:
    return standardise_mode(mode, force_exec=force_exec) & 0o700


def standardise_extended_mode(ext_mode: int) -> int:
    file_type = stat.S_IFMT(ext_mode)
    return file_type | standardise_mode(ext_mode, force_exec=stat.S_ISDIR(file_type))


def is_extended_mode_dir(ext_mode: int) -> bool:
    return stat.S_ISDIR(ext_mode)



import logging
import typing as typ

from matomo_dl.bundle.extract import extract_zip_file
from matomo_dl.bundle.info import BuildInformation
from matomo_dl.distribution.lock import VersionedPluginLock
from matomo_dl.errors import DownloadHashMismatch
from matomo_dl.lock.plugin.versioned import (
    get_cache_key,
    get_plugin_data,
    plugin_request,
)
from matomo_dl.progress import progressbar
from matomo_dl.session import SessionStore

logger = logging.getLogger(__name__)


def extract_plugins(
    session: SessionStore, license_key: typ.Optional[str], build: BuildInformation
):
    plugin_locks = build.lockfile.plugin_locks
    with progressbar(plugin_locks.items(), label="Extracting plugins") as bar:
        for name, plugin_lock in bar:
            if isinstance(plugin_lock, VersionedPluginLock):
                extract_versioned_plugin(session, license_key, build, name, plugin_lock)
            else:
                raise ValueError()


def extract_versioned_plugin(
    session: SessionStore,
    license_key: typ.Optional[str],
    build: BuildInformation,
    name: str,
    lock: VersionedPluginLock,
):
    plugin = get_plugin_data(session, license_key, name)
    plugin.assert_downloadable()  # We perform a license check every build.
    cache_key = get_cache_key(name, lock.version)

    data = session.retrieve_cache_data(cache_key, lock.hash)
    if not data:
        r = plugin_request(session, lock.link, license_key=license_key)
        data = r.content
        assert data is not None
        data_hash = session.store_cache_data(cache_key, data)
        if lock.hash != data_hash:
            raise DownloadHashMismatch()

    latest_mtime = extract_zip_file(
        data, build.folder / "plugins" / plugin.name, root=lock.extraction_root
    )
    assert latest_mtime is not None
    build.add_source_time(latest_mtime)



#!/usr/bin/env python3
import codecs
import os
import typing as typ

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
about: typ.Dict[str, typ.Any] = {}


with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()


with open(os.path.join(here, "matomo_dl", "__version__.py")) as v:
    exec(v.read(), about)  # noqa: S102


required = [
    "attrs~=18.2",
    "beautifulsoup4~=4.6",
    "cattrs==0.9.0",
    "click~=7.0",
    "click-log",
    "colorama==0.4.0",
    "lxml~=4.2",
    "packaging~=18.0",
    "requests~=2.20",
    "toml==0.10.0",
]


setup(
    name="matomo_dl",
    version=about["__version__"],
    description=".",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Opal Symes",
    author_email="opal@catalyst.net.nz",
    packages=find_packages(include=["matomo_dl", "matomo_dl.*"]),
    entry_points={
        "console_scripts": [
            "matomo-dl=matomo_dl.__main__:cli",
            "matomo_dl=matomo_dl.__main__:cli",
        ]
    },
    python_requires=">=3.6",
    install_requires=required,
    extras_require={"cache": ["requests_cache==0.4.13"]},
    include_package_data=True,
    license="GPLv3+",
)



import logging
import pathlib

import click
import click_log

from matomo_dl import __version__
from matomo_dl.bundle import build_release
from matomo_dl.distribution.load_save import (
    diff_lockfiles,
    load_from_distribution_path,
    write_lockfile_using_distribution_path,
)
from matomo_dl.errors import MatomoError
from matomo_dl.lock.general import build_lock
from matomo_dl.session import (
    CACHE_LEVEL_NAMES,
    DEFAULT_CACHE_LEVEL,
    SessionStore,
    create_session,
)

logger = logging.getLogger(__name__)
click_log.basic_config()
cache_level_choices = sorted(
    set(
        tuple(map(str, CACHE_LEVEL_NAMES.keys()))
        + tuple(map(str, CACHE_LEVEL_NAMES.values()))
    )
)


@click.group()
@click.version_option(version=__version__)
@click_log.simple_verbosity_option(logging.root)
@click.option(
    "--cache",
    "--cache-dir",
    "-C",
    "cache_dir",
    type=click.Path(exists=False, file_okay=False),
    show_envvar=True,
)
@click.option(
    "--cache-level",
    "cache_level",
    default=str(DEFAULT_CACHE_LEVEL),
    type=click.Choice(cache_level_choices),
    show_envvar=True,
)
@click.option("--clear", "cache_clear", is_flag=True, default=False)
@click.pass_context
def cli(ctx, cache_dir: str, cache_level: str, cache_clear: bool):
    ctx.ensure_object(dict)

    ctx.obj["session"] = create_session(
        cache_dir=cache_dir, level=cache_level, clear=cache_clear
    )
    pass


@cli.command()
@click.option("--diff/--no-diff", "-d/ ", "diff", default=True)
@click.option("--dry", "dry", default=False, is_flag=True)
@click.argument(
    "distribution_file",
    default="./distribution.toml",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False),
)
@click.pass_context
def update(ctx, distribution_file, dry, diff):
    session = ctx.obj["session"]
    assert isinstance(session, SessionStore)
    distribution_file = pathlib.Path(distribution_file)
    dist, lock = load_from_distribution_path(distribution_file)
    try:
        new_lock = build_lock(session, dist, lock)
    except MatomoError as e:
        click.echo(
            "🛑 "
            + click.style("Error: ", fg="red", bold=True)
            + click.style(str(e), fg="red")
            + " 🛑"
        )
        return ctx.exit(2)
    if new_lock != lock:
        if diff:
            diff_lockfiles(lock, new_lock)
        if dry:
            click.secho("! File has changes !")
            if not diff:
                click.secho("Run with '--diff' to see the changes")
        else:
            write_lockfile_using_distribution_path(distribution_file, new_lock)
            click.secho("✨ Written lock file changes ✨", fg="green", bold=True)
    else:
        click.secho("✨ Already up to date ✨", fg="green")


@cli.command()
@click.option(
    "--fail-if-updates/--no-fail-if-updates", "-f/ ", "fail_if_updates", default=False
)
@click.option(
    "--update/--no-update", "--sync/ ", "-u/-U", "update_locks", default=False
)
@click.option(
    "--output",
    "-o",
    "output_file",
    default="./matomo.tar.gz",
    type=click.Path(exists=False, resolve_path=True, dir_okay=False),
)
@click.argument(
    "distribution_file",
    default="./distribution.toml",
    type=click.Path(exists=True, resolve_path=True, dir_okay=False),
)
@click.pass_context
def build(ctx, distribution_file, output_file, update_locks, fail_if_updates):
    update_kws = {}
    if update_locks is True:
        update_kws["diff"] = True
        update_kws["dry"] = False
    if fail_if_updates is True:
        update_kws["diff"] = True
        update_kws["dry"] = True
    if update_kws:
        click.secho("🔄 Checking for updates before building 🔄")
        ctx.invoke(update, distribution_file=distribution_file, **update_kws)

    click.secho("  Building your distribution  ")
    session = ctx.obj["session"]
    assert isinstance(session, SessionStore)
    distribution_file = pathlib.Path(distribution_file)
    dist, lock = load_from_distribution_path(distribution_file)
    if not lock:
        click.secho(
            "⛔ The distribution file hasn't been locked ⛔", fg="yellow", bold=True
        )
        click.echo("Add '--sync' to create one as part of the build.")
        ctx.exit(1)
    if lock.distribution_hash != dist.versioning_hash:
        click.secho("⛔ The distribution file has changed ⛔", fg="yellow", bold=True)
        click.echo("Cowardly refusing to build from an outdated lock file.")
        click.echo("Add '--sync' to perform an update.")
        ctx.exit(1)
    try:
        build_release(session, dist, lock, pathlib.Path(output_file))
        click.secho("🎉 Built your distribution 🎉", fg="green")
    except MatomoError as e:
        click.echo(
            "💥 "
            + click.style("Error: ", fg="red", bold=True)
            + click.style(str(e), fg="red")
            + " 💥"
        )
        return ctx.exit(2)


if __name__ == "__main__":
    cli(
        auto_envvar_prefix="MATOMO_DL",
        token_normalize_func=lambda s: str(s).strip().lower(),
    )



import logging
import re
import typing as typ
from urllib.parse import urljoin

import bs4
import requests

from matomo_dl.distribution.lock import MatomoLock
from matomo_dl.distribution.version import Version
from matomo_dl.gpg import GpgVerifier, KeyImportError, VerificationError
from matomo_dl.lock import get_zip_extraction_root
from matomo_dl.session import SessionStore

logger = logging.getLogger(__name__)
API_URL = "https://api.matomo.org"
BUILDS_URL = "https://builds.matomo.org"
VERSION_REGEX = re.compile(r".*\/matomo-([0-9]+.*)\.((?:zip|tar\.gz)(?:\.asc)?)$")


def get_cache_key(version: str) -> str:
    return f"matomo-{version}-zip"


def sync_matomo_lock(
    session: SessionStore,
    version_spec: Version,
    existing_lock: typ.Optional[MatomoLock],
) -> MatomoLock:
    logger.info(f"Resolving version spec: {version_spec}")
    version = resolve_matomo_version_spec(session, version_spec)
    if existing_lock and version == existing_lock.version:
        return existing_lock
    logger.info(f"Downloading matomo version {version_spec}")
    cache_key = get_cache_key(version)
    url, data = get_matomo_version(session, version)
    logger.info("Determining extraction root")
    base_path = get_zip_extraction_root(data, "piwik.php")
    if not base_path:
        logger.error("Cannot determine how to extract Matomo!")
        raise ValueError("")
    data_hash = session.store_cache_data(cache_key, data)
    lock = MatomoLock(
        version=version, link=url, hash=data_hash, extraction_root=base_path
    )
    logger.info(f"Stored cache data. Lock: {lock}")
    return lock


def get_matomo_version(
    session: requests.Session, version: str
) -> typ.Tuple[str, bytes]:
    dl_url = f"{BUILDS_URL}/matomo-{version}.zip"
    asc_url = f"{BUILDS_URL}/matomo-{version}.zip.asc"
    logger.info(f"Downloading Matomo release {version}")
    r = session.get(dl_url)
    zip_file = r.content
    r = session.get(asc_url)
    r.raise_for_status()
    zip_file_sig = r.content

    with GpgVerifier() as verifier:
        try:
            verifier.load_fingerprint("0x814E346FA01A20DBB04B6807B5DBD5925590A237")
            verifier.verify(zip_file, zip_file_sig)
        except KeyImportError:
            logger.error("Unable to import the Matomo release keys.")
            raise
        except VerificationError:
            logger.error("Signature does not match file.")
            raise
    return dl_url, zip_file


def resolve_matomo_version_spec(
    session: requests.Session, version_spec: Version
) -> str:
    logger.info(f"Getting latest matomo version from API")
    latest = get_latest_matomo_version(session)
    if version_spec.choose_version([latest]):
        logger.info(f"Latest version({latest}) matches. Using it.")
        return latest
    else:
        logger.info(f"Latest version({latest}) doesn't match, checking all versions")
        all_versions = get_all_matomo_versions(session)
        logger.debug(f"All versions are:{all_versions}")
        version = version_spec.choose_version(set(all_versions))
        logger.info(f"Chosen version: {version}")
        if version:
            return version
        else:
            raise ValueError("No supported versions")


def get_latest_matomo_version(session: requests.Session) -> str:
    resp = session.get(f"{API_URL}/1.0/getLatestVersion/")
    resp.raise_for_status()
    return resp.text.strip()


def get_all_matomo_versions(session: requests.Session) -> typ.Collection[str]:
    resp = session.get(BUILDS_URL)
    resp.raise_for_status()
    base_url = resp.url
    logger.info("Got version listing, parsing with BeautifulSoup")
    soup = bs4.BeautifulSoup(resp.text, "lxml")
    versions = set()
    for a in soup.find_all("a"):
        if "href" not in a.attrs or "matomo" not in a.attrs["href"]:
            continue
        full_link = urljoin(base_url, a.attrs["href"])
        match = VERSION_REGEX.match(full_link)
        logger.debug("Examined link(matches: {match}): {full_link!r}")
        if match:
            versions.add(match.group(1))
    return set(versions)



import logging
import typing as typ

from matomo_dl.bundle.info import BuildInformation
from matomo_dl.call_tree import OrderedCall, order_calls_by_dependencies
from matomo_dl.progress import progressbar

logger = logging.getLogger(__name__)


CustomisationCallable = typ.Callable[[BuildInformation], None]
CustomisationCollection = typ.Collection[OrderedCall]


BASE_CALLS: typ.FrozenSet[OrderedCall] = frozenset(
    {
        OrderedCall("ALL", requires=["CONFIG", "PLUGINS", "FILES"]),
        OrderedCall("CONFIG", requires=["PLUGINS"]),
        OrderedCall("PLUGINS", affects=["FILES"]),
        OrderedCall("FILES"),
    }
)

BASE_CALL_NAMES = frozenset(c.name for c in BASE_CALLS)


def apply_customisations(build: BuildInformation) -> None:
    tree_inputs = set(BASE_CALLS)
    for call in set(build.customisations.get_customisation_functions()):
        tree_inputs.add(call.normalise(BASE_CALL_NAMES))
    calls = order_calls_by_dependencies(tree_inputs)
    with progressbar(calls, label="Applying customisations") as bar:
        for call in bar:
            try:
                call.call(build)
            except Exception as e:
                raise ValueError(f"Call to customisation {call!r} failed") from e



import tarfile
import typing as typ
import zipfile
from io import BytesIO


def get_zip_extraction_root(zip_data, root_file) -> typ.Optional[str]:
    if root_file[0:2] == "./":
        root_file = root_file[2:]
    root_file = root_file.lstrip("/")
    root = None
    with zipfile.ZipFile(BytesIO(zip_data)) as zipf:
        for name in zipf.namelist():
            root_folder, file, _ = name.rpartition(root_file)
            if file == root_file and root_folder[-1] == "/":
                if root is None:
                    root = root_folder
                elif len(root) > len(root_file):
                    root = root_folder
    return root


def get_tar_extraction_root(tar_data, root_file) -> typ.Optional[str]:
    if root_file[0:2] == "./":
        root_file = root_file[2:]
    root_file = root_file.lstrip("/")
    root = None
    with tarfile.open(fileobj=BytesIO(tar_data)) as tar:
        for name in tar.getnames():
            root_folder, file, _ = name.rpartition(root_file)
            if file == root_file and root_folder[-1] == "/":
                if root is None:
                    root = root_file
                elif len(root) > len(root_file):
                    root = root_file
    return root



import logging
import pathlib
import typing as typ
from datetime import timedelta
from types import MappingProxyType

from .caching_store import CachingSessionStore
from .store import SessionStore

logger = logging.getLogger(__name__)
CACHE_LEVEL_NAMES = MappingProxyType(
    {"none": 0, "locks": 1, "memory": 2, "persist": 3, "persistant": 3, "persistent": 3}
)

if CachingSessionStore is None:
    SUPPORTED_CACHE_LEVEL_NAMES = frozenset(("none", "locks"))
else:
    SUPPORTED_CACHE_LEVEL_NAMES = frozenset(("none", "locks", "memory", "persist"))

SUPPORTED_CACHE_LEVELS = frozenset(
    CACHE_LEVEL_NAMES[name] for name in SUPPORTED_CACHE_LEVEL_NAMES
)
MAX_SUPPORTED_CACHE_LEVEL = max(SUPPORTED_CACHE_LEVELS)
DEFAULT_CACHE_LEVEL = min(2, MAX_SUPPORTED_CACHE_LEVEL)


def create_session(
    cache_dir: typ.Union[pathlib.Path, str, None],
    level: typ.Union[str, int, float, None],
    clear: bool = False,
) -> SessionStore:
    cache_level = standardise_level(level)
    del level  # Prevent bugs if we accidentally reuse `level`
    if not cache_dir or cache_level == 0:
        return SessionStore(cache_dir=None)
    else:
        cache_dir = pathlib.Path(cache_dir)
        cache_dir.resolve(strict=False)
        if clear:
            remove_dir(cache_dir)
        cache_dir.mkdir(parents=True, exist_ok=True)
        if cache_level == CACHE_LEVEL_NAMES["memory"]:
            return create_cached_session_store(cache_dir=cache_dir, backend="memory")
        elif cache_level == CACHE_LEVEL_NAMES["persistent"]:
            return create_cached_session_store(cache_dir=cache_dir, backend="sqlite")
        else:
            return SessionStore(cache_dir=cache_dir)


def create_cached_session_store(cache_dir: pathlib.Path, backend: str) -> SessionStore:
    if not CachingSessionStore:
        logger.warning(
            "CachingSessionStore is unavaliable due to a "
            "lack of the underlying library."
        )
        return SessionStore(cache_dir=cache_dir)
    common = {
        "allowable_methods": ["GET", "POST"],
        "old_data_on_error": False,
        "expire_after": timedelta(days=7),
    }
    if backend == "memory":
        return CachingSessionStore(cache_dir=cache_dir, backend="memory", **common)
    else:
        return CachingSessionStore(
            cache_dir=cache_dir,
            cache_name=str(cache_dir / "request-cache"),  # requests_cache argument
            backend_options={"fast_save": True},
            backend="sqlite",
            **common,
        )


def standardise_level(level: typ.Union[str, int, float, None]) -> int:
    if level is None:
        return DEFAULT_CACHE_LEVEL
    elif str(level) in CACHE_LEVEL_NAMES:
        cache_level = CACHE_LEVEL_NAMES[str(level)]
    else:
        cache_level = int(level)
    if cache_level < 0:
        cache_level = 0
    elif cache_level > max(CACHE_LEVEL_NAMES.values()):
        logger.warning(
            f"The given level {level}(mapped to {cache_level}) is outside "
            f"the supported range. Clamping to {MAX_SUPPORTED_CACHE_LEVEL}"
        )
        cache_level = MAX_SUPPORTED_CACHE_LEVEL
    if cache_level not in SUPPORTED_CACHE_LEVELS:
        logger.warning(
            f"The given level {level!r}(mapped to {cache_level}) is not supported, "
            f"likely due to missing libraries. Setting to {DEFAULT_CACHE_LEVEL}"
        )
        cache_level = DEFAULT_CACHE_LEVEL
    assert 0 <= cache_level <= MAX_SUPPORTED_CACHE_LEVEL
    return cache_level


def remove_dir(input_path: pathlib.Path):
    if not input_path.exists():
        return
    if input_path.is_symlink() or input_path.is_file():
        input_path.unlink()
        return
    to_explore = [input_path]
    while to_explore:
        curr = to_explore[-1]
        old_len = len(to_explore)
        assert not curr.is_symlink() and curr.is_dir()
        for path in curr.iterdir():
            if path.is_symlink() or path.is_file():
                path.unlink()
            if path.is_dir():
                to_explore.append(path)
        if old_len == len(to_explore):
            curr.rmdir()
            to_explore.pop()



import logging
import re

from matomo_dl.bundle.info import BuildInformation

logger = logging.getLogger(__name__)
AUTOLOAD_CLASSMAP_ENTRY_RE = re.compile(
    r'^.*? => \$(vendorDir|baseDir) \. ([\'"])(.*?)\2,$'
)
AUTOLOAD_STATIC_ENTRY_RE = re.compile(
    r'^.*? => __DIR__ \. ([\'"])/(.*?)\1 \. ([\'"])(.*?)\3,$'
)


def regenerate_autoload_classmap(build: BuildInformation) -> None:
    folder = build.folder
    bases = {"vendorDir": folder / "vendor", "baseDir": folder}
    autoload = folder / "vendor/composer/autoload_classmap.php"
    autoload_file = iter(autoload.read_text().splitlines(keepends=False))
    prefix = []
    content = []
    postfix = []
    for line in autoload_file:
        prefix.append(line)
        if "return array(" in line:
            break
    for line in autoload_file:
        match = AUTOLOAD_CLASSMAP_ENTRY_RE.fullmatch(line)
        if not match:
            postfix.append(line)
            break
        baseType, fname = match.group(1, 3)
        file = bases[baseType] / f".{fname}"
        if not file.exists() or not file.is_file():
            logger.debug(
                f"autoload_classmap.php Removing {baseType!r}, {fname!r} => ({file.exists()}, {file.is_file()}) {file!r} "
            )
            # Remove missing entries
            continue
        content.append(line)
    postfix.extend(autoload_file)
    autoload.write_text("\n".join(prefix + sorted(content) + postfix))


def regenerate_autoload_static(build: BuildInformation) -> None:
    folder = build.folder
    autoload = folder / "vendor/composer/autoload_static.php"
    base = autoload.parent
    autoload_file = iter(autoload.read_text().splitlines(keepends=False))
    prefix = []
    content = []
    postfix = []
    for line in autoload_file:
        prefix.append(line)
        if "public static $classMap = array (" in line:
            break
    for line in autoload_file:
        match = AUTOLOAD_STATIC_ENTRY_RE.fullmatch(line)
        if not match:
            postfix.append(line)
            break
        cdPart, filePart = match.group(2, 4)
        file = base / f"{cdPart}{filePart}"
        if not file.exists() or not file.is_file():
            logger.debug(
                f"autoload_static.php Removing {cdPart!r}{filePart!r} => ({file.exists()}, {file.is_file()}) {file!r} "
            )
            # Remove missing entries
            continue
        content.append(line)
    postfix.extend(autoload_file)
    autoload.write_text("\n".join(prefix + sorted(content) + postfix))
    pass



import typing as typ

import attr

from matomo_dl.bundle.customisation import (
    CustomisationCollection,
    autoload,
    config,
    manifest,
    remove,
)
from matomo_dl.call_tree import OrderedCall

oc_from_c = OrderedCall.from_callable


def filter_customisations(
    *fns: typ.Tuple[bool, OrderedCall]
) -> CustomisationCollection:
    out_fns = []
    for include, fn in fns:
        if include:
            out_fns.append(fn)
    return out_fns


class Customisation:
    def get_customisation_functions(self) -> CustomisationCollection:
        raise NotImplementedError()


@attr.s
class ManifestCustomisation(Customisation):

    regenerate: bool = attr.ib(default=True)

    def get_customisation_functions(self) -> CustomisationCollection:
        return filter_customisations(
            (
                self.regenerate,
                oc_from_c(manifest.regenerate_manifest, requires=["ALL"]),
            ),
            (
                self.regenerate,
                oc_from_c(autoload.regenerate_autoload_classmap, requires=["ALL"]),
            ),
            (
                self.regenerate,
                oc_from_c(autoload.regenerate_autoload_static, requires=["ALL"]),
            ),
        )


@attr.s
class RemoveCustomisation(Customisation):

    build_support: bool = attr.ib(default=False)
    documentation: bool = attr.ib(default=False)
    example_plugins: bool = attr.ib(default=False)
    vendored_extras: bool = attr.ib(default=False)

    def get_customisation_functions(self) -> CustomisationCollection:
        return filter_customisations(
            (
                self.build_support,
                oc_from_c(remove.remove_build_support, affects=["FILES"]),
            ),
            (
                self.documentation,
                oc_from_c(remove.remove_documentation, affects=["FILES"]),
            ),
            (
                self.example_plugins,
                oc_from_c(remove.remove_example_plugins, affects=["FILES"]),
            ),
            (
                self.vendored_extras,
                oc_from_c(remove.remove_vendored_extras, affects=["FILES"]),
            ),
        )


@attr.s
class UpdateCustomisation(Customisation):

    cacert: bool = attr.ib(default=False)

    def get_customisation_functions(self) -> CustomisationCollection:
        return filter_customisations()


@attr.s
class Customisations(Customisation):
    manifest: typ.Optional[ManifestCustomisation] = attr.ib(default=None)
    remove: typ.Optional[RemoveCustomisation] = attr.ib(default=None)
    update: typ.Optional[UpdateCustomisation] = attr.ib(default=None)

    def get_customisation_functions(self) -> CustomisationCollection:
        fns: typ.List[OrderedCall] = [
            OrderedCall.from_callable(config.update_plugins_list, requires=["FILES"])
        ]
        if self.manifest is not None:
            fns.extend(self.manifest.get_customisation_functions())
        if self.remove is not None:
            fns.extend(self.remove.get_customisation_functions())
        if self.update is not None:
            fns.extend(self.update.get_customisation_functions())
        return fns



import pathlib
import typing as typ

import click

from matomo_dl.colored_diff import ColoredDiffer
from matomo_dl.distribution.file import DistributionFile
from matomo_dl.distribution.lock import DistributionLockFile
from .file import stringify_distribution_file, unstringify_distribution_file
from .lock import stringify_distribution_lock, unstringify_distribution_lock


def lockfile_path_from_distribution(distribution_file: pathlib.Path) -> pathlib.Path:
    distribution_lock_file = distribution_file.with_suffix(
        ".lock" + distribution_file.suffix
    )
    return distribution_lock_file


def normfile_path_from_distribution(distribution_file: pathlib.Path) -> pathlib.Path:
    distribution_lock_file = distribution_file.with_suffix(
        ".norm" + distribution_file.suffix
    )
    return distribution_lock_file


def load_from_distribution_path(
    distribution_file: pathlib.Path
) -> typ.Tuple[DistributionFile, typ.Optional[DistributionLockFile]]:
    distribution_lockfile = lockfile_path_from_distribution(distribution_file)

    dist = unstringify_distribution_file(
        distribution_file.parent, distribution_file.read_text()
    )
    normfile_path_from_distribution(distribution_file).write_text(
        stringify_distribution_file(dist)
    )
    try:
        lock = unstringify_distribution_lock(distribution_lockfile.read_text())
    except Exception:
        lock = None
    return dist, lock


def write_lockfile_using_distribution_path(
    distribution_file: pathlib.Path, locks: DistributionLockFile
):
    distribution_lockfile = lockfile_path_from_distribution(distribution_file)
    distribution_lockfile.write_text(stringify_distribution_lock(locks))


def diff_lockfiles(
    old: typ.Optional[DistributionLockFile], new: DistributionLockFile, quieter=False
):
    if old == new:
        if not quieter:
            click.secho("No changes")
        return
    if old is None:
        old_txt = ""
    else:
        old_txt = stringify_distribution_lock(old)
    new_txt = stringify_distribution_lock(new)
    # We should have a diff at this point; if old != new and old_txt == new_txt then we have a bug
    assert old_txt != new_txt
    if old and old.distribution_hash != new.distribution_hash:
        if not quieter:
            click.secho(
                "# Note: these files were built from different distribution configurations",
                dim=True,
            )
    click.secho("--- Old")
    click.secho("+++ New")
    for line in ColoredDiffer().compare(old_txt, new_txt):
        click.echo(line)



import random
import shutil
import subprocess  # noqa: S404 -- subprocess usage is safe
import tempfile
import typing as typ


class GpgError(Exception):
    pass


class KeyImportError(GpgError):
    pass


class VerificationError(GpgError):
    pass


class GpgVerifier:

    keyservers = frozenset(("keys.gnupg.net", "keyserver.ubuntu.com", "pgp.mit.edu"))

    tmp_folder: typ.Optional[tempfile.TemporaryDirectory] = None

    def __enter__(self) -> "GpgVerifier":
        if self.tmp_folder is None:
            self.tmp_folder = tempfile.TemporaryDirectory()
            self.tmp_folder.__enter__()
        return self

    def __exit__(self, ex_type, value, traceback):
        if self.tmp_folder is not None:
            # Occasional race condition may cause files in the tem directory to
            # disappear during the removal. Just do the best we can
            name = self.tmp_folder.name
            try:
                self.tmp_folder.__exit__(ex_type, value, traceback)
            except IOError:
                try:
                    shutil.rmtree(name, ignore_errors=True)
                except IOError:
                    pass
        self.tmp_folder = None

    def get_tmp_folder(self) -> str:
        if self.tmp_folder:
            return self.tmp_folder.name
        else:
            return self.__enter__().get_tmp_folder()

    def gpg_call(self, *args: str, check=True, **kwargs: typ.Any):
        args = (
            "gpg",
            "--batch",
            "--no-default-keyring",
            "--keyring",
            self.get_tmp_folder() + "/keyring",
            "--homedir",
            self.get_tmp_folder(),
        ) + args
        kwargs.setdefault("input", "")
        return subprocess.run(  # noqa: S603 -- assume callers are safe
            args, check=check, **kwargs
        )

    def load_fingerprint(self, fingerprint: str):
        err = None
        for _attempt in range(3):
            key_server = random.choice(  # noqa: S311 -- psuedo-random is fine
                list(self.keyservers)
            )
            try:
                self.gpg_call("--keyserver", key_server, "--recv-keys", fingerprint)
                return
            except subprocess.CalledProcessError as e:
                err = e
        raise KeyImportError() from err

    def load_public_key(self, key_content: typ.Union[str, bytes]):
        try:
            self.gpg_call("--import", "-", input=key_content)
        except subprocess.CalledProcessError as e:
            raise KeyImportError() from e

    def verify(self, data: bytes, signature: bytes):
        with tempfile.NamedTemporaryFile("wb") as data_file:
            data_file.write(data)
            data_file.flush()
            try:
                self.gpg_call("--verify", "-", data_file.name, input=signature)
            except subprocess.CalledProcessError as e:
                raise VerificationError() from e
